/****************************************************************************
** Meta object code from reading C++ file 'gui_qt6.hpp'
**
** Created by: The Qt Meta Object Compiler version 69 (Qt 6.11.1)
**
** WARNING! All changes made in this file will be lost!
*****************************************************************************/

#include "../../../../guvcview/guvcview/gui_qt6.hpp"
#include <QtGui/qtextcursor.h>
#include <QtGui/qscreen.h>
#include <QtCore/qmetatype.h>

#include <QtCore/qtmochelpers.h>

#include <memory>


#include <QtCore/qxptype_traits.h>
#if !defined(Q_MOC_OUTPUT_REVISION)
#error "The header file 'gui_qt6.hpp' doesn't include <QObject>."
#elif Q_MOC_OUTPUT_REVISION != 69
#error "This file was generated using the moc from 6.11.1. It"
#error "cannot be used with the include files from this version of Qt."
#error "(The moc has changed too much.)"
#endif

#ifndef Q_CONSTINIT
#define Q_CONSTINIT
#endif

QT_WARNING_PUSH
QT_WARNING_DISABLE_DEPRECATED
QT_WARNING_DISABLE_GCC("-Wuseless-cast")
namespace {
struct qt_meta_tag_ZN10MainWindowE_t {};
} // unnamed namespace

template <> constexpr inline auto MainWindow::qt_create_metaobjectdata<qt_meta_tag_ZN10MainWindowE_t>()
{
    namespace QMC = QtMocConstants;
    QtMocHelpers::StringRefStorage qt_stringData {
        "MainWindow",
        "capture_video_clicked",
        "",
        "capture_image_clicked",
        "set_statusbar_message",
        "message",
        "quit_button_clicked",
        "slider_value_changed",
        "value",
        "spin_value_changed",
        "button_PanTilt1_clicked",
        "button_PanTilt2_clicked",
        "button_clicked",
        "pan_tilt_step_changed",
        "combo_changed",
        "index",
        "bayer_pix_ord_changed",
        "check_changed",
        "state",
        "autofocus_changed",
        "setfocus_clicked",
        "int64_button_clicked",
        "string_button_clicked",
        "bitmask_button_clicked",
        "devices_changed",
        "frame_rate_changed",
        "resolution_changed",
        "format_changed",
        "render_fx_filter_changed",
        "render_osd_changed",
        "meet4k_background_mode_changed",
        "meet4k_camera_angle_changed",
        "meet4k_background_changed",
        "meet4k_ColorBG_changed",
        "meet4k_blur_level_changed",
        "meet4k_hdr_mode_changed",
        "meet4k_ae_mode_changed",
        "meet4k_br_mode_changed",
        "meet4k_nr_mode_changed",
        "h264_rate_control_mode_changed",
        "h264_TemporalScaleMode_changed",
        "h264_SpatialScaleMode_changed",
        "h264_FrameInterval_changed",
        "h264_commit_button_clicked",
        "h264_reset_button_clicked",
        "audio_api_changed",
        "audio_devices_changed",
        "audio_samplerate_changed",
        "audio_channels_changed",
        "audio_latency_changed",
        "audio_fx_filter_changed",
        "control_defaults_clicked",
        "load_save_profile_clicked",
        "menu_camera_button_clicked",
        "photo_file_clicked",
        "photo_suffix_clicked",
        "video_file_clicked",
        "video_suffix_clicked",
        "video_codec_clicked",
        "video_codec_properties",
        "audio_codec_clicked",
        "audio_codec_properties",
        "check_device_events",
        "check_control_events"
    };

    QtMocHelpers::UintData qt_methods {
        // Slot 'capture_video_clicked'
        QtMocHelpers::SlotData<void()>(1, 2, QMC::AccessPublic, QMetaType::Void),
        // Slot 'capture_image_clicked'
        QtMocHelpers::SlotData<void()>(3, 2, QMC::AccessPublic, QMetaType::Void),
        // Slot 'set_statusbar_message'
        QtMocHelpers::SlotData<void(QString)>(4, 2, QMC::AccessPublic, QMetaType::Void, {{
            { QMetaType::QString, 5 },
        }}),
        // Slot 'quit_button_clicked'
        QtMocHelpers::SlotData<void()>(6, 2, QMC::AccessPrivate, QMetaType::Void),
        // Slot 'slider_value_changed'
        QtMocHelpers::SlotData<void(int)>(7, 2, QMC::AccessPrivate, QMetaType::Void, {{
            { QMetaType::Int, 8 },
        }}),
        // Slot 'spin_value_changed'
        QtMocHelpers::SlotData<void(int)>(9, 2, QMC::AccessPrivate, QMetaType::Void, {{
            { QMetaType::Int, 8 },
        }}),
        // Slot 'button_PanTilt1_clicked'
        QtMocHelpers::SlotData<void()>(10, 2, QMC::AccessPrivate, QMetaType::Void),
        // Slot 'button_PanTilt2_clicked'
        QtMocHelpers::SlotData<void()>(11, 2, QMC::AccessPrivate, QMetaType::Void),
        // Slot 'button_clicked'
        QtMocHelpers::SlotData<void()>(12, 2, QMC::AccessPrivate, QMetaType::Void),
        // Slot 'pan_tilt_step_changed'
        QtMocHelpers::SlotData<void(int)>(13, 2, QMC::AccessPrivate, QMetaType::Void, {{
            { QMetaType::Int, 8 },
        }}),
        // Slot 'combo_changed'
        QtMocHelpers::SlotData<void(int)>(14, 2, QMC::AccessPrivate, QMetaType::Void, {{
            { QMetaType::Int, 15 },
        }}),
        // Slot 'bayer_pix_ord_changed'
        QtMocHelpers::SlotData<void(int)>(16, 2, QMC::AccessPrivate, QMetaType::Void, {{
            { QMetaType::Int, 15 },
        }}),
        // Slot 'check_changed'
        QtMocHelpers::SlotData<void(int)>(17, 2, QMC::AccessPrivate, QMetaType::Void, {{
            { QMetaType::Int, 18 },
        }}),
        // Slot 'autofocus_changed'
        QtMocHelpers::SlotData<void(int)>(19, 2, QMC::AccessPrivate, QMetaType::Void, {{
            { QMetaType::Int, 18 },
        }}),
        // Slot 'setfocus_clicked'
        QtMocHelpers::SlotData<void()>(20, 2, QMC::AccessPrivate, QMetaType::Void),
        // Slot 'int64_button_clicked'
        QtMocHelpers::SlotData<void()>(21, 2, QMC::AccessPrivate, QMetaType::Void),
        // Slot 'string_button_clicked'
        QtMocHelpers::SlotData<void()>(22, 2, QMC::AccessPrivate, QMetaType::Void),
        // Slot 'bitmask_button_clicked'
        QtMocHelpers::SlotData<void()>(23, 2, QMC::AccessPrivate, QMetaType::Void),
        // Slot 'devices_changed'
        QtMocHelpers::SlotData<void(int)>(24, 2, QMC::AccessPrivate, QMetaType::Void, {{
            { QMetaType::Int, 15 },
        }}),
        // Slot 'frame_rate_changed'
        QtMocHelpers::SlotData<void(int)>(25, 2, QMC::AccessPrivate, QMetaType::Void, {{
            { QMetaType::Int, 15 },
        }}),
        // Slot 'resolution_changed'
        QtMocHelpers::SlotData<void(int)>(26, 2, QMC::AccessPrivate, QMetaType::Void, {{
            { QMetaType::Int, 15 },
        }}),
        // Slot 'format_changed'
        QtMocHelpers::SlotData<void(int)>(27, 2, QMC::AccessPrivate, QMetaType::Void, {{
            { QMetaType::Int, 15 },
        }}),
        // Slot 'render_fx_filter_changed'
        QtMocHelpers::SlotData<void(int)>(28, 2, QMC::AccessPrivate, QMetaType::Void, {{
            { QMetaType::Int, 18 },
        }}),
        // Slot 'render_osd_changed'
        QtMocHelpers::SlotData<void(int)>(29, 2, QMC::AccessPrivate, QMetaType::Void, {{
            { QMetaType::Int, 18 },
        }}),
        // Slot 'meet4k_background_mode_changed'
        QtMocHelpers::SlotData<void(int)>(30, 2, QMC::AccessPrivate, QMetaType::Void, {{
            { QMetaType::Int, 15 },
        }}),
        // Slot 'meet4k_camera_angle_changed'
        QtMocHelpers::SlotData<void(int)>(31, 2, QMC::AccessPrivate, QMetaType::Void, {{
            { QMetaType::Int, 15 },
        }}),
        // Slot 'meet4k_background_changed'
        QtMocHelpers::SlotData<void(int)>(32, 2, QMC::AccessPrivate, QMetaType::Void, {{
            { QMetaType::Int, 15 },
        }}),
        // Slot 'meet4k_ColorBG_changed'
        QtMocHelpers::SlotData<void(int)>(33, 2, QMC::AccessPrivate, QMetaType::Void, {{
            { QMetaType::Int, 15 },
        }}),
        // Slot 'meet4k_blur_level_changed'
        QtMocHelpers::SlotData<void(int)>(34, 2, QMC::AccessPrivate, QMetaType::Void, {{
            { QMetaType::Int, 8 },
        }}),
        // Slot 'meet4k_hdr_mode_changed'
        QtMocHelpers::SlotData<void(int)>(35, 2, QMC::AccessPrivate, QMetaType::Void, {{
            { QMetaType::Int, 8 },
        }}),
        // Slot 'meet4k_ae_mode_changed'
        QtMocHelpers::SlotData<void(int)>(36, 2, QMC::AccessPrivate, QMetaType::Void, {{
            { QMetaType::Int, 8 },
        }}),
        // Slot 'meet4k_br_mode_changed'
        QtMocHelpers::SlotData<void(int)>(37, 2, QMC::AccessPrivate, QMetaType::Void, {{
            { QMetaType::Int, 8 },
        }}),
        // Slot 'meet4k_nr_mode_changed'
        QtMocHelpers::SlotData<void(int)>(38, 2, QMC::AccessPrivate, QMetaType::Void, {{
            { QMetaType::Int, 8 },
        }}),
        // Slot 'h264_rate_control_mode_changed'
        QtMocHelpers::SlotData<void(int)>(39, 2, QMC::AccessPrivate, QMetaType::Void, {{
            { QMetaType::Int, 15 },
        }}),
        // Slot 'h264_TemporalScaleMode_changed'
        QtMocHelpers::SlotData<void(int)>(40, 2, QMC::AccessPrivate, QMetaType::Void, {{
            { QMetaType::Int, 8 },
        }}),
        // Slot 'h264_SpatialScaleMode_changed'
        QtMocHelpers::SlotData<void(int)>(41, 2, QMC::AccessPrivate, QMetaType::Void, {{
            { QMetaType::Int, 8 },
        }}),
        // Slot 'h264_FrameInterval_changed'
        QtMocHelpers::SlotData<void(int)>(42, 2, QMC::AccessPrivate, QMetaType::Void, {{
            { QMetaType::Int, 8 },
        }}),
        // Slot 'h264_commit_button_clicked'
        QtMocHelpers::SlotData<void()>(43, 2, QMC::AccessPrivate, QMetaType::Void),
        // Slot 'h264_reset_button_clicked'
        QtMocHelpers::SlotData<void()>(44, 2, QMC::AccessPrivate, QMetaType::Void),
        // Slot 'audio_api_changed'
        QtMocHelpers::SlotData<void(int)>(45, 2, QMC::AccessPrivate, QMetaType::Void, {{
            { QMetaType::Int, 15 },
        }}),
        // Slot 'audio_devices_changed'
        QtMocHelpers::SlotData<void(int)>(46, 2, QMC::AccessPrivate, QMetaType::Void, {{
            { QMetaType::Int, 15 },
        }}),
        // Slot 'audio_samplerate_changed'
        QtMocHelpers::SlotData<void(int)>(47, 2, QMC::AccessPrivate, QMetaType::Void, {{
            { QMetaType::Int, 15 },
        }}),
        // Slot 'audio_channels_changed'
        QtMocHelpers::SlotData<void(int)>(48, 2, QMC::AccessPrivate, QMetaType::Void, {{
            { QMetaType::Int, 15 },
        }}),
        // Slot 'audio_latency_changed'
        QtMocHelpers::SlotData<void(double)>(49, 2, QMC::AccessPrivate, QMetaType::Void, {{
            { QMetaType::Double, 8 },
        }}),
        // Slot 'audio_fx_filter_changed'
        QtMocHelpers::SlotData<void(int)>(50, 2, QMC::AccessPrivate, QMetaType::Void, {{
            { QMetaType::Int, 18 },
        }}),
        // Slot 'control_defaults_clicked'
        QtMocHelpers::SlotData<void()>(51, 2, QMC::AccessPrivate, QMetaType::Void),
        // Slot 'load_save_profile_clicked'
        QtMocHelpers::SlotData<void()>(52, 2, QMC::AccessPrivate, QMetaType::Void),
        // Slot 'menu_camera_button_clicked'
        QtMocHelpers::SlotData<void()>(53, 2, QMC::AccessPrivate, QMetaType::Void),
        // Slot 'photo_file_clicked'
        QtMocHelpers::SlotData<void()>(54, 2, QMC::AccessPrivate, QMetaType::Void),
        // Slot 'photo_suffix_clicked'
        QtMocHelpers::SlotData<void()>(55, 2, QMC::AccessPrivate, QMetaType::Void),
        // Slot 'video_file_clicked'
        QtMocHelpers::SlotData<void()>(56, 2, QMC::AccessPrivate, QMetaType::Void),
        // Slot 'video_suffix_clicked'
        QtMocHelpers::SlotData<void()>(57, 2, QMC::AccessPrivate, QMetaType::Void),
        // Slot 'video_codec_clicked'
        QtMocHelpers::SlotData<void()>(58, 2, QMC::AccessPrivate, QMetaType::Void),
        // Slot 'video_codec_properties'
        QtMocHelpers::SlotData<void()>(59, 2, QMC::AccessPrivate, QMetaType::Void),
        // Slot 'audio_codec_clicked'
        QtMocHelpers::SlotData<void()>(60, 2, QMC::AccessPrivate, QMetaType::Void),
        // Slot 'audio_codec_properties'
        QtMocHelpers::SlotData<void()>(61, 2, QMC::AccessPrivate, QMetaType::Void),
        // Slot 'check_device_events'
        QtMocHelpers::SlotData<void()>(62, 2, QMC::AccessPrivate, QMetaType::Void),
        // Slot 'check_control_events'
        QtMocHelpers::SlotData<void()>(63, 2, QMC::AccessPrivate, QMetaType::Void),
    };
    QtMocHelpers::UintData qt_properties {
    };
    QtMocHelpers::UintData qt_enums {
    };
    return QtMocHelpers::metaObjectData<MainWindow, qt_meta_tag_ZN10MainWindowE_t>(QMC::MetaObjectFlag{}, qt_stringData,
            qt_methods, qt_properties, qt_enums);
}
Q_CONSTINIT const QMetaObject MainWindow::staticMetaObject = { {
    QMetaObject::SuperData::link<QMainWindow::staticMetaObject>(),
    qt_staticMetaObjectStaticContent<qt_meta_tag_ZN10MainWindowE_t>.stringdata,
    qt_staticMetaObjectStaticContent<qt_meta_tag_ZN10MainWindowE_t>.data,
    qt_static_metacall,
    nullptr,
    qt_staticMetaObjectRelocatingContent<qt_meta_tag_ZN10MainWindowE_t>.metaTypes,
    nullptr
} };

void MainWindow::qt_static_metacall(QObject *_o, QMetaObject::Call _c, int _id, void **_a)
{
    auto *_t = static_cast<MainWindow *>(_o);
    if (_c == QMetaObject::InvokeMetaMethod) {
        switch (_id) {
        case 0: _t->capture_video_clicked(); break;
        case 1: _t->capture_image_clicked(); break;
        case 2: _t->set_statusbar_message((*reinterpret_cast<std::add_pointer_t<QString>>(_a[1]))); break;
        case 3: _t->quit_button_clicked(); break;
        case 4: _t->slider_value_changed((*reinterpret_cast<std::add_pointer_t<int>>(_a[1]))); break;
        case 5: _t->spin_value_changed((*reinterpret_cast<std::add_pointer_t<int>>(_a[1]))); break;
        case 6: _t->button_PanTilt1_clicked(); break;
        case 7: _t->button_PanTilt2_clicked(); break;
        case 8: _t->button_clicked(); break;
        case 9: _t->pan_tilt_step_changed((*reinterpret_cast<std::add_pointer_t<int>>(_a[1]))); break;
        case 10: _t->combo_changed((*reinterpret_cast<std::add_pointer_t<int>>(_a[1]))); break;
        case 11: _t->bayer_pix_ord_changed((*reinterpret_cast<std::add_pointer_t<int>>(_a[1]))); break;
        case 12: _t->check_changed((*reinterpret_cast<std::add_pointer_t<int>>(_a[1]))); break;
        case 13: _t->autofocus_changed((*reinterpret_cast<std::add_pointer_t<int>>(_a[1]))); break;
        case 14: _t->setfocus_clicked(); break;
        case 15: _t->int64_button_clicked(); break;
        case 16: _t->string_button_clicked(); break;
        case 17: _t->bitmask_button_clicked(); break;
        case 18: _t->devices_changed((*reinterpret_cast<std::add_pointer_t<int>>(_a[1]))); break;
        case 19: _t->frame_rate_changed((*reinterpret_cast<std::add_pointer_t<int>>(_a[1]))); break;
        case 20: _t->resolution_changed((*reinterpret_cast<std::add_pointer_t<int>>(_a[1]))); break;
        case 21: _t->format_changed((*reinterpret_cast<std::add_pointer_t<int>>(_a[1]))); break;
        case 22: _t->render_fx_filter_changed((*reinterpret_cast<std::add_pointer_t<int>>(_a[1]))); break;
        case 23: _t->render_osd_changed((*reinterpret_cast<std::add_pointer_t<int>>(_a[1]))); break;
        case 24: _t->meet4k_background_mode_changed((*reinterpret_cast<std::add_pointer_t<int>>(_a[1]))); break;
        case 25: _t->meet4k_camera_angle_changed((*reinterpret_cast<std::add_pointer_t<int>>(_a[1]))); break;
        case 26: _t->meet4k_background_changed((*reinterpret_cast<std::add_pointer_t<int>>(_a[1]))); break;
        case 27: _t->meet4k_ColorBG_changed((*reinterpret_cast<std::add_pointer_t<int>>(_a[1]))); break;
        case 28: _t->meet4k_blur_level_changed((*reinterpret_cast<std::add_pointer_t<int>>(_a[1]))); break;
        case 29: _t->meet4k_hdr_mode_changed((*reinterpret_cast<std::add_pointer_t<int>>(_a[1]))); break;
        case 30: _t->meet4k_ae_mode_changed((*reinterpret_cast<std::add_pointer_t<int>>(_a[1]))); break;
        case 31: _t->meet4k_br_mode_changed((*reinterpret_cast<std::add_pointer_t<int>>(_a[1]))); break;
        case 32: _t->meet4k_nr_mode_changed((*reinterpret_cast<std::add_pointer_t<int>>(_a[1]))); break;
        case 33: _t->h264_rate_control_mode_changed((*reinterpret_cast<std::add_pointer_t<int>>(_a[1]))); break;
        case 34: _t->h264_TemporalScaleMode_changed((*reinterpret_cast<std::add_pointer_t<int>>(_a[1]))); break;
        case 35: _t->h264_SpatialScaleMode_changed((*reinterpret_cast<std::add_pointer_t<int>>(_a[1]))); break;
        case 36: _t->h264_FrameInterval_changed((*reinterpret_cast<std::add_pointer_t<int>>(_a[1]))); break;
        case 37: _t->h264_commit_button_clicked(); break;
        case 38: _t->h264_reset_button_clicked(); break;
        case 39: _t->audio_api_changed((*reinterpret_cast<std::add_pointer_t<int>>(_a[1]))); break;
        case 40: _t->audio_devices_changed((*reinterpret_cast<std::add_pointer_t<int>>(_a[1]))); break;
        case 41: _t->audio_samplerate_changed((*reinterpret_cast<std::add_pointer_t<int>>(_a[1]))); break;
        case 42: _t->audio_channels_changed((*reinterpret_cast<std::add_pointer_t<int>>(_a[1]))); break;
        case 43: _t->audio_latency_changed((*reinterpret_cast<std::add_pointer_t<double>>(_a[1]))); break;
        case 44: _t->audio_fx_filter_changed((*reinterpret_cast<std::add_pointer_t<int>>(_a[1]))); break;
        case 45: _t->control_defaults_clicked(); break;
        case 46: _t->load_save_profile_clicked(); break;
        case 47: _t->menu_camera_button_clicked(); break;
        case 48: _t->photo_file_clicked(); break;
        case 49: _t->photo_suffix_clicked(); break;
        case 50: _t->video_file_clicked(); break;
        case 51: _t->video_suffix_clicked(); break;
        case 52: _t->video_codec_clicked(); break;
        case 53: _t->video_codec_properties(); break;
        case 54: _t->audio_codec_clicked(); break;
        case 55: _t->audio_codec_properties(); break;
        case 56: _t->check_device_events(); break;
        case 57: _t->check_control_events(); break;
        default: ;
        }
    }
}

const QMetaObject *MainWindow::metaObject() const
{
    return QObject::d_ptr->metaObject ? QObject::d_ptr->dynamicMetaObject() : &staticMetaObject;
}

void *MainWindow::qt_metacast(const char *_clname)
{
    if (!_clname) return nullptr;
    if (!strcmp(_clname, qt_staticMetaObjectStaticContent<qt_meta_tag_ZN10MainWindowE_t>.strings))
        return static_cast<void*>(this);
    return QMainWindow::qt_metacast(_clname);
}

int MainWindow::qt_metacall(QMetaObject::Call _c, int _id, void **_a)
{
    _id = QMainWindow::qt_metacall(_c, _id, _a);
    if (_id < 0)
        return _id;
    if (_c == QMetaObject::InvokeMetaMethod) {
        if (_id < 58)
            qt_static_metacall(this, _c, _id, _a);
        _id -= 58;
    }
    if (_c == QMetaObject::RegisterMethodArgumentMetaType) {
        if (_id < 58)
            *reinterpret_cast<QMetaType *>(_a[0]) = QMetaType();
        _id -= 58;
    }
    return _id;
}
QT_WARNING_POP
